#include <random>
#undef min
#undef max
#include <algorithm> 
#include <cmath>

ImVec2 toImVec2(const Vector2& vec) {
    return ImVec2(vec.x, vec.y);
}

struct Particle {
    Vector3 position;
    Vector3 velocity;
    float lifetime;
    std::chrono::steady_clock::time_point startTime;
    ImVec4 color;
};

class ParticleManager {
public:
    void addSuperParticle(const Vector3& position, const ImVec4& color, int particleCount = 20) {
        for (int i = 0; i < particleCount; ++i) {
            Particle particle;
            particle.position = position;
            particle.color = color;

            // Calculate direction evenly around a sphere
            float theta = (2.0f * M_PI * i) / particleCount; // Azimuthal angle
            float phi = static_cast<float>(rand()) / RAND_MAX * M_PI; // Polar angle
            float speed = static_cast<float>(rand()) / RAND_MAX * 0.09f + 0.01f; // Random speed between 0.01 and 0.1

            particle.velocity.x = cos(phi) * cos(theta) * speed;
            particle.velocity.y = cos(phi) * sin(theta) * speed;
            particle.velocity.z = sin(phi) * speed;

            particle.lifetime = 1.0f;
            particle.startTime = std::chrono::steady_clock::now();
            particles.push_back(particle);
        }
    }

    void updateAndRenderParticles() {
        auto now = std::chrono::steady_clock::now();
        particles.erase(
            std::remove_if(particles.begin(), particles.end(), [&](const Particle& particle) {
                return std::chrono::duration<float>(now - particle.startTime).count() > particle.lifetime;
                }),
            particles.end()
        );

        ImDrawList* draw_list = ImGui::GetBackgroundDrawList();

        for (auto& particle : particles) {
            float age = std::chrono::duration<float>(now - particle.startTime).count();
            float t = age / particle.lifetime;

            particle.position += particle.velocity; // Vector3 addition

            Vector2 screenPos = calc::world_to_screen(particle.position);

            ImVec2 imScreenPos = toImVec2(screenPos);

            if (imScreenPos.x >= 0 && imScreenPos.y >= 0) { // Ensure the position is valid on screen
                int glowRadius = 2;
                int glowAlphaStep = 30;
                int maxAlpha = static_cast<int>((1.0f - t) * 255);

                for (int r = glowRadius; r > 0; --r) {
                    float factor = static_cast<float>(r) / glowRadius;
                    ImVec4 glowColor = interpolateColor(ImVec4(particle.color.x, particle.color.y, particle.color.z, 1.0f), ImVec4(0, 0, 0, 0), factor);

                    ImColor imGlowColor(glowColor.x, glowColor.y, glowColor.z, glowColor.w);
                    int alpha = std::max(0, maxAlpha - (glowRadius - r) * glowAlphaStep);

                    draw_list->AddCircleFilled(imScreenPos, static_cast<float>(r), IM_COL32(imGlowColor.Value.x, imGlowColor.Value.y, imGlowColor.Value.z, alpha));
                }

                draw_list->AddCircleFilled(imScreenPos, 3.0f, IM_COL32(particle.color.x, particle.color.y, particle.color.z, maxAlpha));
            }
        }
    }
private:
    std::vector<Particle> particles;

    ImVec4 interpolateColor(const ImVec4& start, const ImVec4& end, float factor) {
        return ImVec4(
            start.x + factor * (end.x - start.x),
            start.y + factor * (end.y - start.y),
            start.z + factor * (end.z - start.z),
            start.w + factor * (end.w - start.w)
        );
    }
};
ParticleManager particleManager;

// Define a structure for notifications
struct Notification {
    std::string message;
    std::chrono::time_point<std::chrono::steady_clock> startTime;
    float duration;
};

// Notification Manager
class NotificationManager {
public:
    void addNotification(const std::string& message, float duration) {
        Notification notification{ message, std::chrono::steady_clock::now(), duration };
        notifications.push_back(notification);
    }

    void updateAndRenderNotifications() {
        auto now = std::chrono::steady_clock::now();
        notifications.erase(
            std::remove_if(notifications.begin(), notifications.end(), [&](const Notification& notification) {
                return std::chrono::duration<float>(now - notification.startTime).count() > notification.duration;
                }),
            notifications.end()
        );

        ImDrawList* draw_list = ImGui::GetBackgroundDrawList();
        const ImVec2 display_size = ImGui::GetIO().DisplaySize;
        const float padding = 5.0f; // Padding around text
        const float margin = 10.0f; // Margin from the screen edges

        ImVec2 start_pos = ImVec2(display_size.x - margin, margin);

        for (size_t i = notifications.size(); i > 0; --i) {
            const Notification& notification = notifications[i - 1];
            ImVec2 text_size = ImGui::CalcTextSize(notification.message.c_str());
            ImVec2 box_min = ImVec2(start_pos.x - text_size.x - padding * 2, start_pos.y);
            ImVec2 box_max = ImVec2(start_pos.x, start_pos.y + text_size.y + padding * 2);

            // Draw filled box with a border
            draw_list->AddRectFilled(box_min, box_max, ImColor(15, 15, 15, 255), 2.f); // Background with transparency
            draw_list->AddRect(box_min, box_max, ImColor(25, 25, 25), 2.f); // Border

            // Draw text
            draw_list->AddText(ImVec2(box_min.x + padding, box_min.y + padding), ImColor(255, 255, 255), notification.message.c_str());

            // Move the starting position for the next notification
            start_pos.y += text_size.y + padding * 2 + 10.0f; // Adjust the spacing between notifications
        }
    }

private:
    std::vector<Notification> notifications;
};

NotificationManager notificationManager;

struct Damage {
    int health_difference;
    Vector3 pos;
    std::chrono::time_point<std::chrono::steady_clock> startTime;
    float duration;
};

class DamageManager {
public:
    void addDamage(const int& health_difference, Vector3 pos, float duration) {
        Damage damage{ health_difference, pos, std::chrono::steady_clock::now(), duration };
        damages.push_back(damage);
    }

    std::vector<Damage> get_damages()
    {
        return damages;
    }

    void updateAndRenderDamageNumbers() {
        auto now = std::chrono::steady_clock::now();
        damages.erase(
            std::remove_if(damages.begin(), damages.end(), [&](const Damage& damage) {
                return std::chrono::duration<float>(now - damage.startTime).count() > damage.duration;
                }),
            damages.end()
        );
    }

private:
    std::vector<Damage> damages;
};

DamageManager damageManager;


namespace features
{
    ImDrawList* draw;

    int calc_text_size(double distance) {
        const int maxTextSize = 20; // Maximum text size
        const int minTextSize = 14; // Minimum text size

        // Calculate text size based on distance
        int textSize = static_cast<int>((maxTextSize - minTextSize) * (1.0 / (1 + distance)) + minTextSize);

        // Ensure textSize doesn't exceed boundaries
        textSize = std::min<int>(maxTextSize, std::max<int>(minTextSize, textSize));

        return textSize;
    }

    int calc_text_sizeB(double distance) {
        const int maxTextSize = 24; // Maximum text size
        const int minTextSize = 18; // Minimum text size

        // Calculate text size based on distance
        int textSize = static_cast<int>((maxTextSize - minTextSize) * (1.0 / (1 + distance)) + minTextSize);

        // Ensure textSize doesn't exceed boundaries
        textSize = std::min<int>(maxTextSize, std::max<int>(minTextSize, textSize));

        return textSize;
    }

    void render_rectangle(int x, int y, int width, int height, overlay::pixel color, int thickness, ImDrawList* draw_list = draw)
    {
        draw_list->AddRect(ImVec2(x, y), ImVec2(x + width, y + height), IM_COL32(color.R, color.G, color.B, color.A), 0.0f, 0, thickness);
    }

    void render_filled_rectangle(int x, int y, int width, int height, overlay::pixel color, ImDrawList* draw_list = draw, float rounding = 0.f)
    {
        draw_list->AddRectFilled(ImVec2(x, y), ImVec2(x + width, y + height), IM_COL32(color.R, color.G, color.B, color.A), rounding, 0);
    }

    void render_filled_triangle(int x, int y, int width, int height, overlay::pixel color, ImDrawList* draw_list = draw)
    {
        draw_list->AddTriangleFilled(ImVec2(x, y), ImVec2(x + width / 2, y - height), ImVec2(x - width / 2, y - height), IM_COL32(color.R, color.G, color.B, color.A));
    }

    void DrawCircle(const Vector3& worldPos, float radius, bool filled, int numPoints, ImColor color, float thickness, ImDrawList* draw_list = draw) {

        if (numPoints >= 200)
            return;
        static ImVec2 points[200];

        float step = 6.2831f / numPoints;
        float theta = 0.f;
        for (int i = 0; i < numPoints; i++, theta += step) {
            Vector3 worldSpace = { worldPos.x + radius * cos(theta), worldPos.y - radius * sin(theta), worldPos.z };
            Vector2 screenSpace = calc::world_to_screen(worldSpace);

            points[i].x = screenSpace.x;
            points[i].y = screenSpace.y;
        }

        if (filled)
            draw_list->AddConvexPolyFilled(points, numPoints, color);
        else
            draw_list->AddPolyline(points, numPoints, color, true, thickness);

    }

    void render_gradient_filled_rectangle(int x, int y, int width, int height, overlay::pixel top_color, overlay::pixel bottom_color, ImDrawList* draw_list = draw)
    {
        ImVec2 top_left(x, y);
        ImVec2 bottom_left(x, y + height);
        ImVec2 top_right(x + width, y);
        ImVec2 bottom_right(x + width, y + height);

        ImU32 col_top = IM_COL32(top_color.R, top_color.G, top_color.B, top_color.A);
        ImU32 col_bottom = IM_COL32(bottom_color.R, bottom_color.G, bottom_color.B, bottom_color.A);

        draw_list->AddRectFilledMultiColor(top_left, bottom_right, col_top, col_top, col_bottom, col_bottom);
    }

    void render_corner_rectangle(int x, int y, int width, int height, overlay::pixel color, int thickness, ImDrawList* draw_list = draw) {
        ImVec2 pos = ImVec2(x, y);
        ImVec2 dim = ImVec2(width, height);
        float cornersizex = dim.x / 4;
        float cornersizey = dim.y / 10;

        draw_list->AddLine(ImVec2(pos), ImVec2(pos.x - cornersizex, pos.y), IM_COL32(color.R, color.G, color.B, color.A), thickness);

        draw_list->AddLine(ImVec2(pos.x - dim.x, pos.y), ImVec2((pos.x - dim.x) + cornersizex, pos.y), IM_COL32(color.R, color.G, color.B, color.A), thickness);

        draw_list->AddLine(ImVec2(pos.x - dim.x, pos.y), ImVec2(pos.x - dim.x, pos.y - cornersizey), IM_COL32(color.R, color.G, color.B, color.A), thickness);

        draw_list->AddLine(ImVec2(pos.x - dim.x, (pos.y - dim.y)), ImVec2(pos.x - dim.x, (pos.y - dim.y) + cornersizey), IM_COL32(color.R, color.G, color.B, color.A), thickness);

        draw_list->AddLine(ImVec2(pos.x - dim.x, (pos.y - dim.y)), ImVec2((pos.x - dim.x) + cornersizex, pos.y - dim.y), IM_COL32(color.R, color.G, color.B, color.A), thickness);

        draw_list->AddLine(ImVec2(pos.x, (pos.y - dim.y)), ImVec2(pos.x - cornersizex, pos.y - dim.y), IM_COL32(color.R, color.G, color.B, color.A), thickness);

        draw_list->AddLine(ImVec2(pos.x, (pos.y - dim.y)), ImVec2(pos.x, (pos.y - dim.y) + cornersizey), IM_COL32(color.R, color.G, color.B, color.A), thickness);

        draw_list->AddLine(ImVec2(pos.x, pos.y), ImVec2(pos.x, pos.y - cornersizey), IM_COL32(color.R, color.G, color.B, color.A), thickness);
    }

    ImFont* esp_font;
    ImFont* krypted_font;
    ImFont* krypted_font_small;

    void render_text(std::string text, int x, int y, float distance, overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color), ImDrawList* draw_list = draw) {
        ImFont* font;
        if (globals::esp::font == 0)
            font = krypted_font;
        else
            font = esp_font;

        //overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color);

        draw_list->AddText(font, calc_text_size(distance), ImVec2(x + 1, y), IM_COL32(0, 0, 0, 255), text.c_str());
        draw_list->AddText(font, calc_text_size(distance), ImVec2(x, y), IM_COL32(text_color.R, text_color.G, text_color.B, 255), text.c_str());
    }

    void render_text_ex(std::string text, int x, int y, float font_size, overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color), ImDrawList* draw_list = draw) {
        ImFont* font;
        if (globals::esp::font == 0)
            font = krypted_font;
        else
            font = esp_font;

        //overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color);

        draw_list->AddText(font, font_size, ImVec2(x + 1, y), IM_COL32(0, 0, 0, 255), text.c_str());
        draw_list->AddText(font, font_size, ImVec2(x, y), IM_COL32(text_color.R, text_color.G, text_color.B, 255), text.c_str());
    }

    void render_textB(std::string text, int x, int y, float distance, overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color), ImDrawList* draw_list = draw) {
        ImFont* font = krypted_font;

        //overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color);

        draw_list->AddText(font, calc_text_sizeB(distance), ImVec2(x + 1, y), IM_COL32(0, 0, 0, 255), text.c_str());
        draw_list->AddText(font, calc_text_sizeB(distance), ImVec2(x, y), IM_COL32(text_color.R, text_color.G, text_color.B, 255), text.c_str());
    }

    // Helper function to split a string by newline characters
    std::vector<std::string> split_lines(const std::string& text) {
        std::vector<std::string> lines;
        std::istringstream stream(text);
        std::string line;
        while (std::getline(stream, line)) {
            lines.push_back(line);
        }
        return lines;
    }

    void render_text_center(std::string text, int x, int y, int width, float distance, ImDrawList* draw_list = draw) {
        ImFont* font;
        if (globals::esp::font == 0)
            font = krypted_font;
        else
            font = esp_font;

        float font_size = calc_text_size(distance);
        if (font_size <= 0.f) {
            // Handle invalid font size
            return;
        }

        std::vector<std::string> lines = split_lines(text);
        float line_height = font->CalcTextSizeA(font_size, FLT_MAX, 0.f, "A").y;

        overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color);

        for (size_t i = 0; i < lines.size(); ++i) {
            float text_size_x = font->CalcTextSizeA(font_size, FLT_MAX, 0.f, lines[i].c_str()).x;

            // Center text horizontally by subtracting half the text width from the x coordinate
            float centered_x = x + width / 2 - text_size_x / 2;
            float line_y = y + i * line_height;

            draw_list->AddText(font, font_size, ImVec2(centered_x + 1, line_y), IM_COL32(0, 0, 0, 255), lines[i].c_str());
            draw_list->AddText(font, font_size, ImVec2(centered_x, line_y), IM_COL32(text_color.R, text_color.G, text_color.B, 255), lines[i].c_str());
        }
    }

    #define RED_TEXT 2743

    void render_text_mul_ex(std::string text, int x, int y, float font_size, ImDrawList* draw_list = draw) {
        ImFont* font;
        if (globals::esp::font == 0)
            font = krypted_font;
        else
            font = esp_font;

        if (font_size <= 0.f) {
            // Handle invalid font size
            return;
        }

        std::vector<std::string> lines = split_lines(text);
        float line_height = font->CalcTextSizeA(font_size, FLT_MAX, 0.f, "A").y;

        overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color);

        for (size_t i = 0; i < lines.size(); ++i) {
            float text_size_x = font->CalcTextSizeA(font_size, FLT_MAX, 0.f, lines[i].c_str()).x;

            // Calculate the position for each line without centering
            float line_x = x;
            float line_y = y + i * line_height;

            overlay::pixel previous_color = text_color;

            // Find and remove RED_TEXT from the current line
            size_t pos = lines[i].find(std::to_string(RED_TEXT));
            if (pos != std::string::npos) {
                // Change text color to red
                text_color.R = 255;
                text_color.G = 0;
                text_color.B = 0;
                text_color.A = 255;

                // Remove the RED_TEXT substring from the line
                lines[i].erase(pos, std::to_string(RED_TEXT).length());
            }
            

            // Add shadow text (black color)
            draw_list->AddText(font, font_size, ImVec2(line_x + 1, line_y), IM_COL32(0, 0, 0, 255), lines[i].c_str());

            // Add main text with defined color
            draw_list->AddText(font, font_size, ImVec2(line_x, line_y), IM_COL32(text_color.R, text_color.G, text_color.B, 255), lines[i].c_str());

            text_color = previous_color;
        }
    }



    void render_text_end(std::string text, int x, int y, float distance, ImDrawList* draw_list = draw) {
        ImFont* font;
        if (globals::esp::font == 0)
            font = krypted_font;
        else
            font = esp_font;
        ImVec2 text_size = font->CalcTextSizeA(calc_text_size(distance), FLT_MAX, 0.0f, text.c_str());

        // Adjust the position to make (x, y) represent the end of the text
        x -= static_cast<int>(text_size.x);

        overlay::pixel text_color = colors::global_to_pixel(globals::esp::text_color);


        draw_list->AddText(font, calc_text_size(distance), ImVec2(x, y), IM_COL32(0, 0, 0, 255), text.c_str());
        draw_list->AddText(font, calc_text_size(distance), ImVec2(x - 1, y - 1), IM_COL32(text_color.R, text_color.G, text_color.B, 255), text.c_str());
    }


    void render_circle(int x, int y, int radius, overlay::pixel color, float thickness)
    {
        draw->AddCircle(ImVec2(x, y), radius, IM_COL32(color.R, color.G, color.B, color.A), 0, thickness);
    }

    void render_filled_circle(int x, int y, int radius, overlay::pixel color)
    {
        draw->AddCircleFilled(ImVec2(x, y), radius, IM_COL32(color.R, color.G, color.B, color.A), 0);
    }

    void render_line(int x1, int y1, int x2, int y2, overlay::pixel color, int thickness, ImDrawList* draw_list = draw)
    {
        draw_list->AddLine(ImVec2(x1, y1), ImVec2(x2, y2), IM_COL32(color.R, color.G, color.B, color.A), thickness);
    }

    void render_hitmarker(int x_center, int y_center, int length, overlay::pixel color, int thickness, ImDrawList* draw_list = draw)
    {
        // Draw four lines forming a hitmarker
        draw_list->AddLine(ImVec2(x_center - length, y_center - length), ImVec2(x_center - length / 2, y_center - length / 2), IM_COL32(color.R, color.G, color.B, color.A), thickness);
        draw_list->AddLine(ImVec2(x_center + length, y_center - length), ImVec2(x_center + length / 2, y_center - length / 2), IM_COL32(color.R, color.G, color.B, color.A), thickness);
        draw_list->AddLine(ImVec2(x_center - length, y_center + length), ImVec2(x_center - length / 2, y_center + length / 2), IM_COL32(color.R, color.G, color.B, color.A), thickness);
        draw_list->AddLine(ImVec2(x_center + length, y_center + length), ImVec2(x_center + length / 2, y_center + length / 2), IM_COL32(color.R, color.G, color.B, color.A), thickness);
    }

    void render_glowing_line(int x1, int y1, int x2, int y2, overlay::pixel color, int thickness, ImDrawList* draw_list = draw)
    {
        // Render multiple lines with decreasing alpha to create a glow effect
        int glow_layers = 10; // Number of layers for the glow effect
        int glow_thickness = thickness;
        float glow_alpha = 0.1f; // Initial alpha for the glow layers

        for (int i = 0; i < glow_layers; ++i)
        {
            // Calculate the current alpha for this layer
            float current_alpha = color.A * (glow_alpha - (i * (glow_alpha / glow_layers)));
            draw_list->AddLine(ImVec2(x1, y1), ImVec2(x2, y2), IM_COL32(color.R, color.G, color.B, (int)current_alpha), glow_thickness);

            // Increase thickness for next layer
            glow_thickness += 2;
        }

        // Render the main line with full opacity and specified thickness
        draw_list->AddLine(ImVec2(x1, y1), ImVec2(x2, y2), IM_COL32(color.R, color.G, color.B, color.A), thickness);
    }

    const float PI = 3.14159265358979323846f;

    float DEG2RAD(float degrees) {
        return degrees * (PI / 180.0f);
    }

    void render_arc(float x, float y, float radius, float min_angle, float max_angle, overlay::pixel color, float thickness, ImDrawList* draw_list = draw) {
        draw_list->PathArcTo(ImVec2(x, y), radius, DEG2RAD(min_angle), DEG2RAD(max_angle), 32);
        draw_list->PathStroke(IM_COL32(color.R, color.G, color.B, color.A), false, thickness);
    }




    Vector3 screen_size = Vector3(screen_width, screen_height, 0);

    void draw_watermark()
    {
        auto now = std::chrono::system_clock::now();
        std::time_t currentTime = std::chrono::system_clock::to_time_t(now);
        std::tm* localTime = std::localtime(&currentTime);
        std::stringstream ss;
        ss << "krypted | " << std::put_time(localTime, "%T")/* << " | FPS: " << static_cast<int>(ImGui::GetIO().Framerate)*/;
        std::string text = ss.str();

        ImVec2 textSize = krypted_font->CalcTextSizeA(12.f, FLT_MAX, 0.f, text.c_str());
        ImVec2 textPos(globals::windows::watermark_position.x, globals::windows::watermark_position.y);

        // Draw black background rectangle
        draw->AddRectFilled(textPos - ImVec2(2, 2), textPos + textSize + ImVec2(2, 2), IM_COL32(0, 0, 0, 255), 4.f);

        // Draw text on top of the black background
        draw->AddText(krypted_font, 12, textPos, IM_COL32(255, 255, 255, 255), text.c_str());
    }

    std::string lock_status_transform(bool b)
    {
        if (b)
            return "Locked";
        else
            return "Unlocked";
    }

    void draw_vehicle_esp()
    {
        std::lock_guard<std::mutex> lock(globals::mutex::vehicle_list_mutex);

        for (const auto& vehicle : vehicle_list::vehicle_list)
        {
            if (globals::esp::vehicle_max_distance < vehicle.distance)
                continue;

            if (vehicle.screen_position.x < 0 || vehicle.screen_position.x > globals::screen.x || vehicle.screen_position.y < 0 || vehicle.screen_position.y > globals::screen.y)
                continue;

            if (globals::toggles::ignore_own_vehicle && vehicle.is_own_vehicle && game::local_ped.in_vehicle())
                continue;

            //render_text(vehicle.vehicle_name, vehicle.screen_position.x, vehicle.screen_position.y, vehicle.distance);



            //if (globals::toggles::draw_healthbar_vehicle && vehicle.health != 0)
            //{

            //    overlay::pixel healthbar_color;
            //    if (globals::toggles::dynamic_healthbar_vehicle)
            //    {
            //        float full_color[] = { 30 / 255, 152 / 255, 39 / 255 };
            //        float half_color[] = { 1.f, 0.65f, 0.f };
            //        float red_color[] = { 1.f, 0.f, 0.f };

            //        if (vehicle.health >= vehicle.max_health / 2)
            //            healthbar_color = colors::global_to_pixel(full_color);

            //        if (vehicle.health < vehicle.max_health / 2)
            //            healthbar_color = colors::global_to_pixel(half_color);

            //        if (vehicle.health <= vehicle.max_health / 4)
            //            healthbar_color = colors::global_to_pixel(red_color);

            //    }
            //    else
            //    {
            //        healthbar_color = colors::global_to_pixel(globals::esp::healthbar_color_vehicle);
            //    }

            //    Vector3 origin_left = vehicle.origin;
            //    Vector3 origin_right = vehicle.origin;

            //    origin_left.x += 0.5f;
            //    origin_right.x -= 0.5f;


            //    Vector2 origin_left_screen = calc::world_to_screen(origin_left);
            //    Vector2 origin_right_screen = calc::world_to_screen(origin_right);

            //    int width = abs(origin_left_screen.x - origin_right_screen.x);

            //    if (width < 20)
            //        width = 20;

            //    render_filled_rectangle(
            //        vehicle.screen_position.x - width / 2,
            //        vehicle.screen_position.y,
            //        width,
            //        7,
            //        { 28, 26, 26, 255 }
            //    );

            //    render_filled_rectangle(
            //        vehicle.screen_position.x - width / 2 + 1,
            //        vehicle.screen_position.y - 1,
            //        width - (width * (vehicle.max_health - vehicle.health) / vehicle.max_health) - 2,
            //        5,
            //        healthbar_color
            //    );
            //}



            if (globals::toggles::draw_lock_status || globals::toggles::draw_vehicle_names || globals::toggles::draw_distance) {

                std::string text;

                if (globals::toggles::draw_vehicle_names)
                    text = vehicle.vehicle_name + "\n";

                if (globals::toggles::draw_vehicle_distance)
                    text = text + "[" + std::to_string((int)vehicle.distance) + "m]\n";

                if (globals::toggles::draw_lock_status)
                    text = text + lock_status_transform(vehicle.is_locked) + "\n";

                if (globals::toggles::draw_driver_status && vehicle.has_driver)
                    text = text + "Has Driver\n";
                if (globals::esp::text_position_vehicle == 0)
                {
                    render_text(
                        text,
                        vehicle.screen_position.x,
                        vehicle.screen_position.y + 20,
                        vehicle.distance
                    );
                }
                else {
                    render_text_center(
                        text,
                        vehicle.screen_position.x,
                        vehicle.screen_position.y + 20,
                        0,
                        vehicle.distance
                    );
                }



            }
        }
    }

    void draw_esp()
    {
        for (const auto& players : entity_list::ped_list)
        {
            if (globals::toggles::ignore_dead_esp && players.health == 0)
                continue;

            if (globals::toggles::ignore_npc && players.is_npc)
                continue;

            if (globals::esp::max_distance < players.distance)
                continue;

            float arrow_width = 5.f;

            if (globals::toggles::draw_oof_indicators)
                render_arc(globals::screen_center.x, globals::screen_center.y, globals::oof_indicator_radius, players.angle - arrow_width, players.angle + arrow_width, colors::global_to_pixel_alpha(globals::esp::oof_indicator_color), 4.f);

            if (players.screen_head.x < 0 || players.screen_head.x > globals::screen.x || players.screen_head.y < 0 || players.screen_head.y > globals::screen.y || players.width > globals::screen.x || players.height > globals::screen.y)
                continue;

            if (players.is_friend && globals::toggles::friend_marker)
            {
                Vector2 screen_marker = calc::world_to_screen({ players.head.x, players.head.y, players.head.z + 0.3f });

                render_filled_triangle(screen_marker.x, screen_marker.y, 10, 5, colors::global_to_pixel_alpha(globals::esp::friend_marker_color));
            }

            bool is_target = players.m_pointer == entity_list::target_ped.m_pointer;

            if (is_target && globals::toggles::target_marker)
            {
                Vector2 screen_marker = calc::world_to_screen({ players.head.x, players.head.y, players.head.z + 0.3f });

                render_filled_triangle(screen_marker.x, screen_marker.y, 10, 5, colors::global_to_pixel_alpha(globals::esp::target_marker_color));
            }

            if (players.is_friend && globals::toggles::ignore_friend_esp)
                continue;
            if (players.is_npc && globals::toggles::ignore_npc)
                continue;

            //Vector2 look_screen = calc::world_to_screen(players.look_line);

            //render_line(players.screen_head.x, players.screen_head.y, look_screen.x, look_screen.y, { 255, 0, 0, 255 }, 1.f);

            if (globals::toggles::draw_hat)
            {
                Vector3 vecHeadPosition = players.head;

                float flStep = (2 * M_PI) / 18.f;
                float flRadius = 0.25;

                for (float rotation = 0; rotation < (M_PI * 2.0); rotation += flStep) {

                    Vector3 pointA(flRadius * cos(rotation) + vecHeadPosition.x, flRadius * sin(rotation) + vecHeadPosition.y, vecHeadPosition.z + 0.03);
                    Vector3 bottomRim(flRadius / 2 * cos(rotation) + vecHeadPosition.x, flRadius / 2.f * sin(rotation) + vecHeadPosition.y, vecHeadPosition.z + 0.15);

                    Vector3 pointB(flRadius * cos(rotation + flStep) + vecHeadPosition.x, flRadius * sin(rotation + flStep) + vecHeadPosition.y, vecHeadPosition.z + 0.03);
                    Vector3 bottomRimPredict(flRadius / 2 * cos(rotation + flStep) + vecHeadPosition.x, flRadius / 2.f * sin(rotation + flStep) + vecHeadPosition.y, vecHeadPosition.z + 0.15);

                    Vector2 topOnScreen, botOnScreen, topOnScreenPredict, botOnScreenPredict;

                    topOnScreen = calc::world_to_screen(pointA);
                    botOnScreen = calc::world_to_screen(bottomRim);
                    topOnScreenPredict = calc::world_to_screen(pointB);
                    botOnScreenPredict = calc::world_to_screen(bottomRimPredict);

                    render_line(topOnScreen.x, topOnScreen.y, botOnScreen.x, botOnScreen.y, colors::global_to_pixel_alpha(globals::esp::hat_color), 1.f);
                    render_line(botOnScreen.x, botOnScreen.y, botOnScreenPredict.x, botOnScreenPredict.y, colors::global_to_pixel_alpha(globals::esp::hat_color), 1.f);
                }
            }



            if (globals::toggles::draw_box)
            {
                overlay::pixel col;

                if (players.is_friend)
                    col = colors::global_to_pixel(globals::esp::friend_box_color);
                else
                    col = colors::global_to_pixel(globals::esp::box_color);
                if (is_target)
                    col = colors::global_to_pixel(globals::esp::target_box_color);

                switch (globals::esp::box_type)
                {
                case 0: // 2d
                    render_rectangle(
                        players.bounds.left,
                        players.bounds.top,
                        players.width,
                        players.height,
                        col,
                        1
                    );

                    break;
                case 1: // 2d corner
                    render_corner_rectangle(
                        players.bounds.left,
                        players.bounds.top,
                        -players.width,
                        -players.height,
                        col,
                        1
                    );
                    break;
                case 2:

                    /*Vector2 front_origin = calc::world_to_screen({ players.origin.x, players.origin.y - 0.5f, players.origin.z });
                    Vector2 back_origin = calc::world_to_screen({ players.origin.x, players.origin.y + 0.5f, players.origin.z });

                    Vector2 front = calc::world_to_screen({ players.head.x, players.head.y - 0.5f, players.head.z });
                    Vector2 back = calc::world_to_screen({ players.head.x, players.head.y + 0.5f, players.head.z });

                    float front_height = front_origin.y - front.y;
                    float back_height = back_origin.y - back.y;

                    float front_width = front_height * 1.8f;
                    float back_width = back_height * 1.8f;

                    render_line(front.x - front_width / 2, front.y, back.x - back_width / 2, back.y, colors::global_to_pixel(globals::esp::box_color), 1);
                    render_line(front.x + front_width / 2, front.y, back.x + back_width / 2, back.y, colors::global_to_pixel(globals::esp::box_color), 1);



                    render_line(front.x - front_width / 2, front.y + front_height, back.x - back_width / 2, back.y + back_height, colors::global_to_pixel(globals::esp::box_color), 1);
                    render_line(front.x + front_width / 2, front.y + front_height, back.x + back_width / 2, back.y + back_height, colors::global_to_pixel(globals::esp::box_color), 1);


                    render_rectangle(
                        front.x - front_width / 2,
                        front.y,
                        front_width,
                        front_height,
                        colors::global_to_pixel(globals::esp::box_color),
                        1
                    );

                    render_rectangle(
                        back.x - back_width / 2,
                        back.y,
                        back_width,
                        back_height,
                        colors::global_to_pixel(globals::esp::box_color),
                        1
                    );*/

                    Vector2 back_origin = calc::world_to_screen({ players.origin.x, players.origin.y - 0.5f, players.origin.z });

                    Vector2 back_head = calc::world_to_screen({ players.head.x, players.head.y - 0.5f, players.head.z });


                    float back_height = (back_origin.y - back_head.y);

                    //float back_width = back_height * 1.8f;


                    Vector2 front_origin = calc::world_to_screen({ players.origin.x, players.origin.y + 0.5f, players.origin.z });

                    Vector2 front_head = calc::world_to_screen({ players.head.x, players.head.y + 0.5f, players.head.z });

                    Vector2 left_origin = calc::world_to_screen({ players.origin.x + 0.5f, players.origin.y , players.origin.z });
                    Vector2 right_origin = calc::world_to_screen({ players.origin.x - 0.5f, players.origin.y , players.origin.z });


                    float front_height = front_origin.y - front_head.y;

                    //float front_width = front_height * 1.8f;

                    render_line(front_origin.x, front_origin.y, back_origin.x, back_origin.y, colors::global_to_pixel(globals::esp::box_color), 1);
                    render_line(left_origin.x, left_origin.y, right_origin.x, right_origin.y, colors::global_to_pixel(globals::esp::box_color), 1);

                    Vector2 left_back = calc::world_to_screen({ players.origin.x + 0.5f, players.origin.y - 0.5f , players.origin.z });
                    Vector2 right_back = calc::world_to_screen({ players.origin.x - 0.5f, players.origin.y - 0.5f , players.origin.z });

                    float back_width = right_back.x - left_back.x;

                    render_line(left_back.x - back_width, left_back.y - back_height / 2, left_back.x + back_width, left_back.y + back_height / 2, colors::global_to_pixel(globals::esp::box_color), 1);
                    render_line(left_back.x, left_back.y - back_height / 2, left_back.x, left_back.y + back_height / 2, colors::global_to_pixel(globals::esp::box_color), 1);


                    break;
                }


            }

            if (globals::toggles::draw_filled_box)
            {
                overlay::pixel col;

                if (players.is_friend)
                    col = colors::global_to_pixel_alpha(globals::esp::friend_filled_box_color);
                else
                    col = colors::global_to_pixel_alpha(globals::esp::filled_box_color);

                if (is_target)
                    col = colors::global_to_pixel_alpha(globals::esp::target_filled_box_color);

                if (globals::esp::filled_box_type == 0)
                {
                    render_filled_rectangle(
                        players.bounds.left,
                        players.bounds.top,
                        players.width,
                        players.height,
                        col
                    );
                }
                else if (globals::esp::filled_box_type == 1) {
                    float invisible_color[] = { 0.f, 0.f, 0.f, 0.f };


                    render_gradient_filled_rectangle(
                        players.bounds.left,
                        players.bounds.top,
                        players.width,
                        players.height,
                        colors::global_to_pixel_alpha(invisible_color),
                        col
                    );
                }
            }



            if (globals::toggles::draw_snaplines)
                render_line(game::local_screen_position.x, game::local_screen_position.y, players.screen_position.x, players.screen_position.y, colors::global_to_pixel(globals::esp::snaplines_color), 2);

            if (globals::toggles::draw_skeleton) {
                overlay::pixel col;

                if (players.is_friend)
                    col = colors::global_to_pixel_alpha(globals::esp::friend_skeleton_color);
                else
                    col = colors::global_to_pixel_alpha(globals::esp::skeleton_color);

                if (is_target)
                    col = colors::global_to_pixel_alpha(globals::esp::target_skeleton_color);

                for (const auto& bones : players.bones) {
                    render_line(bones.first.x, bones.first.y, bones.second.x, bones.second.y, col, 2);
                }
            }

            if (globals::toggles::draw_joints) {
                overlay::pixel col;

                if (players.is_friend)
                    col = colors::global_to_pixel_alpha(globals::esp::friend_joints_color);
                else
                    col = colors::global_to_pixel_alpha(globals::esp::joints_color);

                if (is_target)
                    col = colors::global_to_pixel_alpha(globals::esp::target_joints_color);

                for (const auto& bones : players.bones) {
                    render_filled_circle(bones.first.x, bones.first.y, 3, col);
                    render_filled_circle(bones.second.x, bones.second.y, 3, col);
                }
            }

            if (globals::toggles::draw_healthbar)
            {
                if (!(globals::toggles::hide_bar_if_zero && players.health == 0))
                {
                    overlay::pixel healthbar_color;
                    if (globals::toggles::dynamic_healthbar)
                    {
                        float full_color[] = { 30 / 255, 152 / 255, 39 / 255 };
                        float half_color[] = { 1.f, 0.65f, 0.f };
                        float red_color[] = { 1.f, 0.f, 0.f };

                        if (players.health >= 100)
                            healthbar_color = colors::global_to_pixel(full_color);

                        if (players.health < 100)
                            healthbar_color = colors::global_to_pixel(half_color);

                        if (players.health <= 50)
                            healthbar_color = colors::global_to_pixel(red_color);
                    }
                    else
                    {
                        healthbar_color = colors::global_to_pixel(globals::esp::healthbar_color);
                    }

                    if (globals::esp::bar_position == 0) // side
                    {
                        //players.bounds.left,
                        //    players.bounds.top,
                        //    players.width,
                        //    players.height,
                        render_filled_rectangle(
                            players.bounds.left - 10 - 1,
                            players.bounds.top + (players.height * (100 - 100) / 100),
                            5,
                            (players.height) + 2,
                            { 28, 26, 26, 255 }
                        );

                        if (players.health != 0)
                        {
                            render_filled_rectangle(
                                players.bounds.left - 10,
                                players.bounds.top + ((players.height) * (200 - players.health) / 200) + 1,
                                3,
                                (players.height) - ((players.height) * (200 - players.health) / 200) - 1,
                                healthbar_color
                            );
                        }

                    }
                    else { // bottom
                        int width = players.width;

                        if (width < 15)
                            width = 15;

                        render_filled_rectangle(
                            players.bounds.left,
                            players.bounds.bottom + 20,
                            width,
                            5,
                            { 28, 26, 26, 255 }
                        );

                        if (players.health != 0)
                        {
                            render_filled_rectangle(
                                players.bounds.left + 1,
                                players.bounds.bottom + 21,
                                width - (width * (200 - players.health) / 200) - 2,
                                3,
                                healthbar_color
                            );
                        }


                    }
                }
            }

            if (globals::toggles::draw_armorbar)
            {
                if (!(globals::toggles::hide_bar_if_zero && players.armor == 0))
                {
                    if (globals::esp::bar_position == 0) // side
                    {
                        render_filled_rectangle(
                            players.bounds.right + 10 - 1,
                            players.bounds.top + (players.height * (100 - 100) / 100),
                            5,
                            (players.height),
                            { 28, 26, 26, 255 }
                        );

                        if (players.armor != 0)
                        {
                            render_filled_rectangle(
                                players.bounds.right + 10,
                                players.bounds.top + ((players.height) * (100 - players.armor) / 100) + 1,
                                3,
                                (players.height) - ((players.height) * (100 - players.armor) / 100) - 1,
                                colors::global_to_pixel(globals::esp::armorbar_color)
                            );
                        }


                    }
                    else { // bottom
                        int width = players.width;

                        if (width < 15)
                            width = 15;

                        render_filled_rectangle(
                            players.bounds.left,
                            players.bounds.bottom + 15,
                            width,
                            5,
                            { 28, 26, 26, 255 }
                        );

                        if (players.armor != 0)
                        {
                            render_filled_rectangle(
                                players.bounds.left + 1,
                                players.bounds.bottom + 16,
                                width - (width * (100 - players.armor) / 100) - 2,
                                3,
                                colors::global_to_pixel(globals::esp::armorbar_color)
                            );
                        }
                    }
                }
            }

            // TEXT

            if ((players.distance) < globals::esp::max_distance_text)
            {
                if (globals::esp::bar_position == 0) // side
                {
                    if (globals::toggles::draw_health) {
                        render_text_end(
                            std::to_string(players.health),
                            players.screen_head.x - (players.width / 2 + 30),
                            players.screen_head.y,
                            players.distance
                        );
                    }

                    if (!(globals::toggles::hide_bar_if_zero && players.armor == 0) && globals::toggles::draw_armor)
                    {
                        render_text(
                            std::to_string(players.armor),
                            players.screen_head.x + (players.width / 2 + 35),
                            players.screen_head.y,
                            players.distance
                        );
                    }
                }
                else { // top
                    if (globals::toggles::draw_health || globals::toggles::draw_armor) {
                        std::string text;

                        if (globals::toggles::draw_health)
                            text = std::to_string(players.health) + "\n";

                        if (!(globals::toggles::hide_bar_if_zero && players.armor == 0) && globals::toggles::draw_armor)
                            text = text + std::to_string(players.armor);

                        render_text(
                            text,
                            players.screen_head.x + (players.width / 2 + 30),
                            players.screen_head.y + players.height + 5,
                            players.distance
                        );
                    }
                }


                if (globals::toggles::draw_weapon_name || globals::toggles::draw_distance || globals::toggles::draw_player_name) {

                    std::string text;

                    if (globals::toggles::draw_player_name && players.player_name != "")
                        text = players.player_name + "\n";

                    if (globals::toggles::draw_weapon_name)
                        text = text + players.weapon_name + "\n";

                    if (globals::toggles::draw_distance)
                        text = text + "[" + std::to_string(players.distance) + "m]\n";

                    if (globals::toggles::draw_net_id && players.net_id != 0)
                        text = text + "ID: " + std::to_string(players.net_id) + "\n";


                    int text_offset = 0;

                    if (globals::esp::bar_position == 1)
                        text_offset = 25;

                    if (globals::esp::text_position == 0)
                    {
                        render_text(
                            text,
                            players.screen_head.x - players.width / 2,
                            players.screen_head.y + players.height * 1.05f + text_offset,
                            players.distance
                        );
                    }
                    else if (globals::esp::text_position == 1)
                    {
                        render_text_center(
                            text,
                            players.screen_head.x - players.width / 2,
                            players.screen_head.y + players.height * 1.05f + text_offset,
                            players.width,
                            players.distance
                        );
                    }


                }
            }
        }
    }

    void draw_fov()
    {
        if (globals::toggles::draw_fov)
            render_circle(screen_width / 2, screen_height / 2, globals::fov, colors::global_to_pixel(globals::fov_color), 2.f);
        if (globals::toggles::draw_filled_fov)
            render_filled_circle(screen_width / 2, screen_height / 2, globals::fov, colors::global_to_pixel_alpha(globals::filled_fov_color));
    }

    inline Vector2 target_bone;

    void draw_target_line()
    {
        std::lock_guard<std::mutex> lock(globals::mutex::target_bone_mutex);

        if (target_bone.x != -1)
            render_line(screen_width / 2, screen_height / 2, target_bone.x, target_bone.y, colors::global_to_pixel(globals::target_line_color), 2);
    }

    void draw_info(std::string s1, std::string s2)
    {
        ImGui::PushStyleColor(ImGuiCol_Text, IM_COL32(255, 255, 255, 153));
        ImGui::Text(s1.c_str());
        ImGui::PopStyleColor();
        ImGui::SameLine();
        ImGui::Text(s2.c_str());
    }

    std::string bool_to_string(bool b)
    {
        if (b)
            return "True";
        else
            return "False";
    }

    void draw_nearby_players_list(bool check_for_invisiblity = false)
    {
        ImGui::PushStyleVar(ImGuiStyleVar_WindowRounding, 8);
        ImGui::PushStyleVar(ImGuiStyleVar_WindowPadding, ImVec2(0, 0));
        ImGui::PushStyleVar(ImGuiStyleVar_WindowMinSize, ImVec2(0, 0));

        ImGui::SetNextWindowPos(ImVec2(10, 250), ImGuiCond_FirstUseEver);
        if (!check_for_invisiblity)
            ImGui::SetNextWindowSize(ImVec2(175, 90), ImGuiCond_FirstUseEver);
        else
            ImGui::SetNextWindowSize(ImVec2(175, 200), ImGuiCond_FirstUseEver);

        std::string window_label = "krypted nearby_players_list";

        if (check_for_invisiblity)
            window_label = "krypted nearby_invisible_players_list";


        if (ImGui::Begin(window_label.c_str(), nullptr, ImGuiWindowFlags_NoDecoration))
        {
            ImGuiWindow* window = ImGui::GetCurrentWindow();
            ImVec2 pos = ImGui::GetWindowPos();

            std::string title = "Nearby Players";

            if (check_for_invisiblity)
                title = "Spectator List";

            ImGui::GetWindowDrawList()->AddRectFilled(pos, ImVec2(pos.x + 175, pos.y + 30), to_vec4(25, 25, 25, custom.m_anim), 8, ImDrawFlags_RoundCornersTop);
            ImGui::GetWindowDrawList()->AddText(pos + ImVec2(10, 8), custom.get_accent_color(custom.m_anim), title.c_str());
            ImGui::SetCursorPos({ 10, 35 });

            std::vector<entity_list::ped_struct> sorted_ped_list = entity_list::ped_list; // Make a copy of the ped_list
            std::sort(sorted_ped_list.begin(), sorted_ped_list.end(), [](const entity_list::ped_struct& a, const entity_list::ped_struct& b) {
                return a.distance < b.distance;
                });

            for (const auto& players : sorted_ped_list) {
                if (check_for_invisiblity && !players.is_invisible)
                    continue;

                if (players.distance < 100.f && players.net_id != 0 && !players.is_self) {
                    ImGui::SetCursorPosX(10);
                    std::string player_label = players.player_name + " [" + std::to_string(players.net_id) + "]";

                    if (players.is_friend)
                        ImGui::PushStyleColor(ImGuiCol_Text, ImVec4(0, 255, 0, 255));

                    ImGui::Text(player_label.c_str());

                    if (players.is_friend)
                        ImGui::PopStyleColor();
                }
            }

            ImGui::SetWindowSize(ImVec2(175, 133));
        }

        ImGui::End();

        ImGui::PopStyleVar(3);
    }


    void draw_spotify_widget()
    {
        ImGui::PushStyleVar(ImGuiStyleVar_WindowRounding, 8);
        ImGui::PushStyleVar(ImGuiStyleVar_WindowPadding, ImVec2(0, 0));
        ImGui::PushStyleVar(ImGuiStyleVar_WindowMinSize, ImVec2(0, 0));

        ImGui::SetNextWindowPos(ImVec2(10, 250), ImGuiCond_FirstUseEver);

        if (ImGui::Begin("krypted spotify widget", nullptr, ImGuiWindowFlags_NoDecoration))
        {
            ImGuiWindow* window = ImGui::GetCurrentWindow();
            ImVec2 pos = ImGui::GetWindowPos();

            std::string title = "Spotify";

            ImGui::GetWindowDrawList()->AddRectFilled(pos, ImVec2(pos.x + 175, pos.y + 30), to_vec4(25, 25, 25, custom.m_anim), 8, ImDrawFlags_RoundCornersTop);
            ImGui::GetWindowDrawList()->AddText(pos + ImVec2(10, 8), custom.get_accent_color(custom.m_anim), title.c_str());
            ImGui::SetCursorPos({ 10, 35 });

            ImGui::Text(spotify::current_song.title.c_str());
            ImGui::SetCursorPosX(10);
            ImGui::PushStyleColor(ImGuiCol_Text, ImVec4(1.f, 1.f, 1.f, 0.6f));
            ImGui::Text(spotify::current_song.author.c_str());
            ImGui::PopStyleColor();

            ImGui::SetWindowSize(ImVec2(175, 80));
            if (spotify::current_song.author == "")
            {
                ImGui::SetWindowSize(ImVec2(175, 60));
            }
        }

        ImGui::End();

        ImGui::PopStyleVar(3);
    }


    void draw_player_info()
    {
        if (!globals::toggles::draw_player_info)
            return;

        ImGui::PushStyleVar(ImGuiStyleVar_WindowRounding, 8);
        ImGui::PushStyleVar(ImGuiStyleVar_WindowPadding, ImVec2(0, 0));
        ImGui::PushStyleVar(ImGuiStyleVar_WindowMinSize, ImVec2(0, 0));

        ImGui::SetNextWindowPos(ImVec2(10, 50), ImGuiCond_FirstUseEver);
        ImGui::SetNextWindowSize(ImVec2(175, 90), ImGuiCond_FirstUseEver);

        if (ImGui::Begin("krypted player_info", nullptr, ImGuiWindowFlags_NoDecoration))
        {
            ImGuiWindow* window = ImGui::GetCurrentWindow();
            ImVec2 pos = ImGui::GetWindowPos();
            ImGui::GetWindowDrawList()->AddRectFilled(pos, ImVec2(pos.x + 175, pos.y + 30), to_vec4(25, 25, 25, custom.m_anim), 8, ImDrawFlags_RoundCornersTop);
            std::string title = entity_list::target_ped.player_name + " [" + std::to_string(entity_list::target_ped.net_id) + "]";
            ImGui::GetWindowDrawList()->AddText(pos + ImVec2(10, 8), custom.get_accent_color(custom.m_anim), title.c_str());
            ImGui::SetCursorPos({ 10, 35 });
            draw_info("Health", std::to_string(entity_list::target_ped.health));
            ImGui::SetCursorPosX(10);
            draw_info("Armor", std::to_string(entity_list::target_ped.armor));
            ImGui::SetCursorPosX(10);
            draw_info("Invisible", bool_to_string(entity_list::target_ped.invisible));
            ImGui::SetCursorPosX(10);
            draw_info("Godmode", bool_to_string(entity_list::target_ped.godmode));
            ImGui::SetCursorPosX(10);
            draw_info("Weapon:", entity_list::target_ped.weapon_name.c_str());
            ImGui::SetCursorPosX(10);
            draw_info("Distance", std::to_string(entity_list::target_ped.distance) + "m");
            ImGui::SetCursorPosX(10);
            ImGui::SetWindowSize(ImVec2(175, 155));

        }

        ImGui::End();

        ImGui::PopStyleVar(3);
    }

    void draw_crosshair()
    {
        render_filled_rectangle(globals::screen_center.x, globals::screen_center.y - 9, 2, 20, colors::global_to_pixel_alpha(globals::crosshair_color));
        render_filled_rectangle(globals::screen_center.x - 9, globals::screen_center.y, 20, 2, colors::global_to_pixel_alpha(globals::crosshair_color));
    }

    void player_looking_warning(int number_of_players, std::string player_name)
    {
        if (!globals::toggles::players_looking_warning)
            return;

        if (number_of_players == 0)
            return;

        std::string text;

        if (number_of_players == 1)
            text = player_name + " is looking at you";  // Note the space before "is"
        else
            text = std::to_string(number_of_players) + " are looking at you";

        ImVec2 text_size = ImGui::CalcTextSize(text.c_str());

        // Define the padding for the background box
        const float padding = 5.0f;

        // Calculate the position and size of the background box
        float screen_width = globals::screen_center.x * 2; // Assuming screen_center is half of the screen width
        float screen_height = globals::screen_center.y * 2; // Assuming screen_center is half of the screen height
        ImVec2 box_pos(screen_width / 2 - text_size.x / 2 - padding, screen_height - text_size.y - padding * 2 - 30);
        ImVec2 box_size(text_size.x + padding * 2, text_size.y + padding * 2);

        ImDrawList* draw_list = ImGui::GetBackgroundDrawList();

        // Draw the background box
        draw_list->AddRectFilled(box_pos, ImVec2(box_pos.x + box_size.x, box_pos.y + box_size.y), ImColor(15, 15, 15, 255));

        // Draw the text
        draw_list->AddText(ImVec2(screen_width / 2 - text_size.x / 2, screen_height - text_size.y - padding - 30), ImColor(255, 255, 255, 255), text.c_str());
    }

    void render_tracers()
    {
        if (!globals::toggles::tracers)
            return;

        std::lock_guard<std::mutex> lock(globals::mutex::tracer_mutex);

        std::vector<features::tracer> tracers = features::tracer_manager.get_tracers();

        for (auto& tracer : tracers)
        {
            Vector2 start_screen = calc::world_to_screen(tracer.start_pos);
            Vector2 end_screen = calc::world_to_screen(tracer.end_pos);

            overlay::pixel tracer_col = colors::global_to_pixel(globals::tracer_color);

            if (globals::tracer_type == 0)
                render_glowing_line(start_screen.x, start_screen.y, end_screen.x, end_screen.y, { tracer_col.B, tracer_col.G, tracer_col.R, static_cast<BYTE>(static_cast<int>(tracer.current_alpha)) }, 3.f);
            else if (globals::tracer_type == 1)
                render_line(start_screen.x, start_screen.y, end_screen.x, end_screen.y, { tracer_col.B, tracer_col.G, tracer_col.R, static_cast<BYTE>(static_cast<int>(tracer.current_alpha)) }, 3.f);

        }
    }

    void render_bullet_impacts()
    {
        if (!globals::toggles::bullet_impact_circle && !globals::toggles::bullet_impact_hitmarker)
            return;

        std::lock_guard<std::mutex> lock(globals::mutex::bullet_impacts_mutex);

        std::vector<features::bullet_impact> bullet_impacts = features::bullet_impacts_manager.get_bullet_impacts();

        for (auto& bullet_impact : bullet_impacts)
        {
            Vector2 end_screen = calc::world_to_screen(bullet_impact.end_pos);

            overlay::pixel bullet_impact_circle_col = colors::global_to_pixel(globals::bullet_impact_circle_color);

            if (globals::toggles::bullet_impact_circle)
                render_circle(end_screen.x, end_screen.y, 5.f, { bullet_impact_circle_col.B, bullet_impact_circle_col.G, bullet_impact_circle_col.R, static_cast<BYTE>(static_cast<int>(bullet_impact.current_alpha)) }, 1.f);

            overlay::pixel bullet_impact_hitmarker_col = colors::global_to_pixel(globals::bullet_impact_hitmarker_color);

            if (globals::toggles::bullet_impact_hitmarker)
                render_hitmarker(end_screen.x, end_screen.y, 8.f, { bullet_impact_hitmarker_col.B, bullet_impact_hitmarker_col.G, bullet_impact_hitmarker_col.R, static_cast<BYTE>(static_cast<int>(bullet_impact.current_alpha)) }, 1.f);
        }
    }

    #include <cmath> 

    void render_ammo_bar()
    {
        //if (!GetAsyncKeyState(0x1) && !GetAsyncKeyState(0x2))
        //    return;

        int target_width = globals::ammobar_width;

        static float current_width = 0.0f; // Current width of the ammo bar for animation

        // Define the animation speed
        const float animation_speed = 1.0f; // Higher value means faster animation

        // Calculate the target width based on the ammo count
        if (features::local_ammo_info.ammo > 0 && features::local_ammo_info.clip_size > 0)
        {
            float ammo_ratio = static_cast<float>(features::local_ammo_info.ammo) / features::local_ammo_info.clip_size;
            target_width = globals::ammobar_width * ammo_ratio;
        }
        else
        {
            target_width = 0;
        }

        // Smoothly transition the current width to the target width
        if (std::abs(current_width - target_width) > 0.1f) // Threshold to prevent jittering
        {
            if (current_width < target_width)
            {
                current_width += animation_speed;
                if (current_width > target_width)
                    current_width = target_width;
            }
            else if (current_width > target_width)
            {
                current_width -= animation_speed;
                if (current_width < target_width)
                    current_width = target_width;
            }
        }

        // Render the background of the ammo bar
        render_filled_rectangle(
            globals::screen_center.x - globals::ammobar_width / 2,
            globals::screen_center.y + 20,
            globals::ammobar_width,
            5,
            { 28, 26, 26, 255 }
        );

        // Render the current ammo bar
        if (current_width > 0)
        {
            render_filled_rectangle(
                globals::screen_center.x - globals::ammobar_width / 2 + 1,
                globals::screen_center.y + 21,
                static_cast<int>(current_width) - 2,
                3,
                colors::global_to_pixel(globals::ammobar_color)
            );
        }




    }

    void draw_crosshair_text()
    {
        std::string text;

        if (globals::toggles::crosshair_text_health)
            text = std::to_string(static_cast<int>(sdk::c_ped::get_health(game::local_player))) + "\n";


        if (globals::toggles::crosshair_text_health)
            text += std::to_string(static_cast<int>(sdk::c_ped::get_armor(game::local_player))) + "\n";

        if (globals::toggles::crosshair_text_target_health && entity_list::target_ped.m_pointer)
            text += std::to_string(RED_TEXT) + std::to_string(static_cast<int>(sdk::c_ped::get_health(entity_list::target_ped.m_pointer))) + "\n";

        if (globals::toggles::crosshair_text_target_armor && entity_list::target_ped.m_pointer)
            text += std::to_string(RED_TEXT) + std::to_string(static_cast<int>(sdk::c_ped::get_armor(entity_list::target_ped.m_pointer))) + "\n";

        render_text_mul_ex(text, globals::screen_center.x - globals::ammobar_width / 2, globals::screen_center.y + 30, 12.f);

        //render_text_ex(std::to_string(static_cast<int>(sdk::c_ped::get_health(game::local_player))),
//    globals::screen_center.x - globals::ammobar_width / 2 + 1,
//    globals::screen_center.y + 30,
//    12.f
//);

//render_text_ex(std::to_string(static_cast<int>(sdk::c_ped::get_armor(game::local_player))),
//    globals::screen_center.x - globals::ammobar_width / 2 + 1,
//    globals::screen_center.y + 40,
//    12.f
//);
    }

    void draw_damage_numbers()
    {
        auto now = std::chrono::steady_clock::now();
        std::vector<Damage> damages = damageManager.get_damages();
        for (size_t i = damages.size(); i > 0; --i) {
            const Damage& damage = damages[i - 1];
            float elapsed_time = std::chrono::duration<float>(now - damage.startTime).count();

            Vector3 new_pos = damage.pos;
            new_pos.y += 0.5f;

            Vector2 screen_pos = calc::world_to_screen(new_pos);

            BYTE alpha = 255;
            if (elapsed_time > 2.0f) {
                float fade_factor = (damage.duration - elapsed_time) / (damage.duration - 2.0f);
                fade_factor = std::fmax(0.0f, fade_factor);
                alpha = static_cast<int>(255 * fade_factor);
            }

            float distance = calc::distance_v3(game::local_position, new_pos);
            ImFont* font = krypted_font;

            overlay::pixel col = colors::global_to_pixel(globals::damage_number_color);

            draw->AddText(font, calc_text_sizeB(distance), ImVec2(screen_pos.x + 1, screen_pos.y), IM_COL32(0, 0, 0, alpha), std::to_string(damage.health_difference).c_str());
            draw->AddText(font, calc_text_sizeB(distance), ImVec2(screen_pos.x, screen_pos.y), IM_COL32(col.R, col.G, col.B, alpha), std::to_string(damage.health_difference).c_str());

        }
    }

    struct Instruction {
        const char* key;
        const char* action;
    };


    // Text is overlapping the box, make box same color as menu, another color for keys bg, make it dependable on freecam target type and check if none | color finished and some chekcs
    // add globals toggles check to function call
    // add mutex so it works properly remember lock has to be renamed to something like lock_target_freecam as lock_target is already used | finished
    // clean up code and their names
    // add more features btw like vehicle rain while looking at a player
    // physgun by holding and putting the vehicle in front of a player and maybe adjusting how far it is from someone with mouse wheel 
    void draw_freecam_instructions()
    {
        ImVec2 display_size = ImGui::GetIO().DisplaySize;

        std::vector<Instruction> instructions = {
            {"W", "Forward"}
        };

        m_freecam_target target = freecam_target;

        switch (target.m_type)
        {
        case PED:
            instructions.push_back({ "E", globals::vehicle_rain[target.m_network_id] ? "Deactivate Vehicle Rain" : "Activate Vehicle Rain" });
            break;
        case VEHICLE:
            instructions.push_back({ "E", "Random Vehicle Color" });
            instructions.push_back({ "G", "Set on Fire" });
            instructions.push_back({ "H", "Teleport into Sky" });
            break;
        case NONE:
            return;
        }

        // Define padding and margin
        float padding = 5.0f;
        float margin = 10.0f;
        float key_box_padding = 2.0f;
        float bg_padding = 10.0f;

        // Calculate the total width and the height of the block of text
        float total_width = 0.0f;
        float max_height = 0.0f;

        std::vector<ImVec2> key_sizes;
        std::vector<ImVec2> action_sizes;

        for (const auto& instr : instructions) {
            ImVec2 key_size = ImGui::CalcTextSize(instr.key);
            ImVec2 action_size = ImGui::CalcTextSize(instr.action);

            key_sizes.push_back(key_size);
            action_sizes.push_back(action_size);

            total_width += key_size.x + action_size.x + 4 * key_box_padding + padding;
            if (key_size.y > max_height) max_height = key_size.y;
            if (action_size.y > max_height) max_height = action_size.y;
        }

        total_width -= padding; // Remove extra padding added in the last loop
        total_width += bg_padding; // Add padding to the right side

        // Calculate the position to draw the text at the bottom right of the screen
        ImVec2 window_pos = ImVec2(display_size.x - total_width - 2 * bg_padding - margin, display_size.y - max_height - 2 * key_box_padding - 2 * bg_padding - margin);

        // Calculate the size of the entire block with background padding
        ImVec2 block_size = ImVec2(total_width + 2 * bg_padding, max_height + 2 * key_box_padding + 2 * bg_padding);

        // Begin a new ImGui window
        ImGui::SetNextWindowPos(window_pos);
        ImGui::SetNextWindowSize(block_size);
        ImGui::Begin("Freecam Instructions", nullptr, ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoScrollWithMouse | ImGuiWindowFlags_NoBackground);

        // Draw the background for the entire block
        ImDrawList* draw_list = ImGui::GetWindowDrawList();
        draw_list->AddRectFilled(window_pos, ImVec2(window_pos.x + block_size.x, window_pos.y + block_size.y), IM_COL32(15, 15, 15, 255)); // Semi-transparent black background

        // Draw the instructions
        float cursor_x = window_pos.x + bg_padding;
        float cursor_y = window_pos.y + bg_padding + 2;

        for (size_t i = 0; i < instructions.size(); ++i) {
            // Draw key box
            ImGui::SetCursorScreenPos(ImVec2(cursor_x, cursor_y - 2));
            ImGui::PushStyleColor(ImGuiCol_Button, IM_COL32(35, 35, 35, 255)); // Light color for key box
            ImGui::PushStyleColor(ImGuiCol_ButtonHovered, IM_COL32(35, 35, 35, 255)); // Light color for key box when hovered
            ImGui::PushStyleColor(ImGuiCol_ButtonActive, IM_COL32(35, 35, 35, 255)); // Light color for key box when active
            ImGui::Button(instructions[i].key);
            ImGui::PopStyleColor(3);

            cursor_x += key_sizes[i].x + 4 * key_box_padding + padding;

            // Draw action text
            ImGui::SetCursorScreenPos(ImVec2(cursor_x, cursor_y));
            ImGui::Text("%s", instructions[i].action);

            cursor_x += action_sizes[i].x + padding;
        }

        // End the ImGui window
        ImGui::End();
    }

    void freecam_marker()
    {
        m_freecam_target target = freecam_target;
        if (target.m_type == NONE)
            return;

        //render_line(target.v_s_origin.x, target.v_s_origin.y, globals::screen_center.x, globals::screen_center.y, { 255, 255, 255, 255 }, 1.f);

        Vector2 screen_marker = calc::world_to_screen({ target.v_origin.x, target.v_origin.y, target.v_origin.z + 0.8f });

        render_filled_triangle(screen_marker.x, screen_marker.y, 20, 10, { 0, 255, 0, 255 });
    }

    void render(ImDrawList* draw_list)
    {
        if (GetForegroundWindow() != game::hwnd && GetForegroundWindow() != game::menu_hwnd)
            return;

        if (!globals::is_logged_in)
            return;

        draw = draw_list;

        std::lock_guard<std::mutex> lock(globals::mutex::entity_list_mutex);

        if (entity_list::target_ped.pixel_distance < 100 || GetForegroundWindow() == game::menu_hwnd)
            draw_player_info();

        if ((GetForegroundWindow() == game::menu_hwnd || GetForegroundWindow() == game::hwnd) && globals::toggles::draw_nearby_players_list)
            draw_nearby_players_list();

        if ((GetForegroundWindow() == game::menu_hwnd || GetForegroundWindow() == game::hwnd) && globals::toggles::spotify_widget)
            draw_spotify_widget();


        if ((GetForegroundWindow() == game::menu_hwnd || GetForegroundWindow() == game::hwnd) && globals::toggles::draw_invisible_players_list)
            draw_nearby_players_list(true);

        if (globals::toggles::draw_watermark)
            draw_watermark();

        //DrawCircle({ 295.256, 192.743, 103.331 }, 1.f, false, 100, {255, 0, 0, 255}, 3.f);

        draw_esp();
        draw_vehicle_esp();
        draw_fov();
        if (globals::toggles::draw_target_line)
            draw_target_line();

        if (globals::toggles::draw_crosshair)
            draw_crosshair();

        notificationManager.updateAndRenderNotifications();
        damageManager.updateAndRenderDamageNumbers();

        if (globals::toggles::hit_particles)
            particleManager.updateAndRenderParticles();

        if (globals::toggles::damage_numbers)
            draw_damage_numbers();

        int players_looking = 0;
        std::string player_name;

        for (const auto& player : entity_list::ped_list) {
            if (player.is_looking_at_you && player.distance < 100 && !player.is_npc && !player.is_self)
            {
                players_looking++;
                player_name = player.player_name;
            }
        }

        player_looking_warning(players_looking, player_name);
        render_tracers();
        render_bullet_impacts();

        draw_crosshair_text();

        if (globals::toggles::ammobar)
            render_ammo_bar();

        //std::lock_guard<std::mutex> lock_freecam(globals::mutex::freecam_mutex);

        //draw_freecam_instructions();
        //freecam_marker();

    }
}
